import subprocess


class CommandLineUtils(object):
	"""
	A class used to run bash commands.

	...

	Methods
	-------
	run_bash_command(command)
		Runs a bash command and returns its output and error
	"""

	@staticmethod
	def run_bash_command(command):
		"""Static method for executing a bash command.

		Parameters
		----------
		command : str
			The bash command to run

		"""

		process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
		output, error = process.communicate()
		return output, error



class NetworkStatisticsServiceException(Exception):
	"""
	Class used to encapsulate top level exceptions.
	"""
	pass

class CommandOutputFormatException(NetworkStatisticsServiceException):
	"""
	Class used to encapsulate exceptions regarding unexpected output format from executed bash commands.
	"""
	pass

class LatencyStatisticsException(NetworkStatisticsServiceException):
	"""
	Class used to encapsulate exceptions regarding unexpected format of the ping command.
	"""
	pass

class SpeedStatisticsException(NetworkStatisticsServiceException):
	"""
	Class used to encapsulate exceptions regarding unexpected format of the speedets cli.
	"""
	pass

class DynamoDbException(NetworkStatisticsServiceException):
	"""
	Class used to encapsulate exceptions regarding DynamoDB APIs.
	"""
	pass

class CommandRunException(NetworkStatisticsServiceException):
	"""
	Class used to encapsulate exceptions regarding executing bash commands.
	"""
	pass


class StringUtils(object):
	""" 
	A class used to hold various util mthods for working with strings.

	...

	Methods
	-------
	remove_empty_strings_from_string_list(string_list)
		returns a new list without the empty strings
	"""

	@staticmethod
	def remove_empty_strings_from_string_list(string_list):
		"""Removes empty strings from a list.

		Parameters
		----------
		string_list : [str]
			A list of strings
		"""

		return list(filter(None, string_list))



class TtlConfig(object):
	""" 
	A class used to pass TTL Configuration to the DynamoDB Statistics Uploader.

	...

	Attributes
	----------
	time_in_months : int
		The time in months an object should live in the DB.
	attribute_name : str
		The name of the attribute where the ttl time should be stored.

	"""


	def __init__(self, time_in_months: int, attribute_name: str):
		"""
		Parameters
		----------
		time_in_months : int
			The time in months an object should live in the DB.
		attribute_name : str
			The name of the attribute where the ttl time should be stored.
			"""

		super().__init__()

		if time_in_months is None or attribute_name is None:
			raise ValueError("None of the TTL Config attributes can be none")

		if time_in_months < 0:
			raise ValueError("Time in months must be a positive number")

		if len(attribute_name) == 0:
			raise ValueError("Attribute name cannot be the empty string")

		self.time_in_months = time_in_months
		self.attribute_name = attribute_name


class DynamoDBConfig(object):
	""" 
	A class used to pass DynamoDB configuration to the Statistics Service.

	...

	Attributes
	----------
	aws_profile : str
		The AWS profile from the credential files to use (default 'default')
	aws_region : str
		The AWS region where the DynamoDB tables reside
	stats_table : str
		The name of the table where your statistics should be uplaoded
	logs_table : str
		The name of the table where error logs should be uplaoded
	ttl_config : TtlConfig
		Time to live config for DynamoDB tables.

	Methods
	-------
	is_valid(check_for_stats)
		Verifies if the given configuration is valid for the upload type
	"""


	def __init__(self, aws_profile = 'default', aws_region = None, stats_table = None, logs_table = None, ttl_config = None):
		"""
		Parameters
		----------
		aws_profile : str
			The AWS profile from the credential files to use (default 'default')
		aws_region : str
			The AWS region where the DynamoDB tables reside
		stats_table : str
			The name of the table where your statistics should be uplaoded
		logs_table : str
			The name of the table where error logs should be uplaoded
		ttl_config : TtlConfig
			Time to live config for DynamoDB tables.
		"""

		super().__init__()
		self.aws_profile = aws_profile
		self.aws_region = aws_region
		self.stats_table = stats_table
		self.logs_table = logs_table
		self.ttl_config = ttl_config
	
	
	def is_valid(self, check_for_stats):
		"""Verifies if the given configuration is valid for the upload type.

		If check_for_stats is True, the configuration is checked for statistics upload. If it 
		is false, the configuration is checked for error logs upload.

		Parameters
		----------
		check_for_stats : bool
			Whether the configuration should be validated for statistc upload
		"""
		
		if self.aws_region is None:
			return False
		if check_for_stats and self.stats_table is None:
			return False
		if not check_for_stats and self.logs_table is None:
			return False
		return True


from NetworkStatisticsService.command_line_utils import CommandLineUtils
from NetworkStatisticsService.latency_statistic_utils import LatencyStatisticUtils
from NetworkStatisticsService.speed_statistic_utils import SpeedStatisticUtils
from NetworkStatisticsService.time_utils import TimeUtils
from NetworkStatisticsService.informal_statistics_uploader import InformalStatisticUploaderInterface
from NetworkStatisticsService.dynamodb_statistics_uploader import DynamoDBStatisticsUploader
from NetworkStatisticsService.exceptions import CommandRunException, DynamoDbException, LatencyStatisticsException, NetworkStatisticsServiceException, SpeedStatisticsException

class StatisticsService(object):
	
	def __init__(self, default_ping_count = 20, stats_uploader = None):
		super().__init__()
		if stats_uploader is not None:
			if not issubclass(stats_uploader, InformalStatisticUploaderInterface):
				raise ValueError("Provided statistics uploader does not implement the InformalStatisticUploaderInterface.")
		
		self.__default_ping_count = default_ping_count
		self.__stats_uploader = stats_uploader

		self.speed_statistics_enabled = True
		self.latency_statistics_enabled = True

		self.stats_upload_enabled = False
		self.logs_upload_enabled = False
		
		self.dynamo_db_config = None
	

	def generate_statistics(self, override_ping_count=None):
		timestamp_dict = TimeUtils.get_current_timestamp()

		try:
			latency_stats = self.__gather_latency_statistics(override_ping_count)
			speed_stats = self.__gather_speed_statistics()  
			full_stats = {**latency_stats, **speed_stats, **timestamp_dict}
			self.__upload_statistics(full_stats)
			return full_stats
		except NetworkStatisticsServiceException as re:
			error_dict = {**{'Error': re}, **TimeUtils.get_current_timestamp()}
			self.__upload_logs(error_dict)
			return error_dict


	def __gather_latency_statistics(self, override_ping_count):
		if self.latency_statistics_enabled:
			ping_count = self.__default_ping_count
			if override_ping_count is not None:
				ping_count = override_ping_count

			try:
				ping_command = "ping www.google.com -c " + str(ping_count)
				statistics_rows = self.__gather_statistics_from_command(ping_command)
				latency_statistics = LatencyStatisticUtils.extract_latency_statistics_from_command_output(statistics_rows)
				return latency_statistics
			except (LatencyStatisticsException, CommandRunException) as re:
				raise NetworkStatisticsServiceException("Latency statistics failed. Nested error: " + str(re))
		else:
			return {}


	def __gather_speed_statistics(self):
		if self.speed_statistics_enabled:
			try:
				statistics_rows = self.__gather_statistics_from_command("speedtest --simple")
				speed_statistics = SpeedStatisticUtils.extract_speed_statistics_from_command_output(statistics_rows)
				return speed_statistics
			except (SpeedStatisticsException, CommandRunException) as re:
				raise NetworkStatisticsServiceException("Speed statistics failed. Nested error: " + str(re))
		else:
			return {}


	def __gather_statistics_from_command(self, command):
		output, error = CommandLineUtils.run_bash_command(command)
		if error is not None:
			raise CommandRunException("Command " + command + " failed with error:\n" + error)
		else:
			output_rows = output.decode().split("\n")
			return list(filter(None, output_rows))[-2:]


	def __upload_statistics(self, stats_dict):
		if self.stats_upload_enabled:
			if self.__stats_uploader is None:
				# Defaults to using dynamoDB.
				if self.dynamo_db_config is None or not self.dynamo_db_config.is_valid(True):
					raise DynamoDbException("Upload enabled and no other provider offered. Dynamo DB configuration is incomplete.\n" + 
									"Please set values to aws_region and dynamodb_stats_table_name.")
				self.__stats_uploader = DynamoDBStatisticsUploader(self.dynamo_db_config)
			self.__stats_uploader.upload_statistics(stats_dict)


	def __upload_logs(self, logs_dict):
		if self.logs_upload_enabled:
			if self.__stats_uploader is None:
				# Defaults to using dynamoDB.
				if self.dynamo_db_config is None or not self.dynamo_db_config.is_valid(False):
					raise DynamoDbException("Upload enabled and no other provider offered. Dynamo DB configuration is incomplete.\n" + 
									"Please set values to aws_region and dynamodb_logs_table_name.")
				self.__stats_uploader = DynamoDBStatisticsUploader(self.dynamo_db_config)
			self.__stats_uploader.upload_statistics(logs_dict)
		else:
			print(logs_dict)




class StatisticUploaderMeta(type):
	"""
	A StatisticUploader metaclass that will be used for Statistic Uploader class creation.
	"""

	def __instancecheck__(cls, instance):
		return cls.__subclasscheck__(type(instance))

	def __subclasscheck__(cls, subclass):
		return (hasattr(subclass, 'upload_statistics') and
				callable(subclass.upload_statistics) and
				hasattr(subclass, 'upload_logs') and
				callable(subclass.upload_logs))


class InformalStatisticUploaderInterface(metaclass=StatisticUploaderMeta):
	"""
	An interface which defines the required methods for the statistic uploader.

	Methods
	-------
	upload_statistics(stats_dict)
		Stub method
	upload_logs(logs_dict)
		Stub method
	"""
	def upload_statistics(self, stats_dict: dict):
		pass

	def upload_logs(self, logs_dict: dict):
		pass
    


from NetworkStatisticsService.exceptions import SpeedStatisticsException

class SpeedStatisticUtils(object):
	""" 
	A class used to extract speed statistics from speedteest command output

	...

	Methods
	-------
	extract_speed_statistics_from_command_output(statistics_rows)
		returns statistics dictionary or raises an error
	"""

	@staticmethod
	def extract_speed_statistics_from_command_output(statistics_rows):
		"""Extracts speed statistics from speedtest command output rows.

		Parameters
		----------
		statistics_rows : [str]
			The rows from the speedtest command output containing the network stats

		Raises
		------
		SpeedStatisticsException
			If any of the rows has unexpected format
		"""

		try: 
			stats = {}
			for row in statistics_rows:
				elems = row.split(":")
				stats[elems[0]] = float(elems[1].strip().split()[0].strip())
			return stats
		except IndexError as ie:
			raise SpeedStatisticsException("An error occured when processing the speed statistics.\nUnderlying error: " + str(ie))



import time


class TimeUtils(object):
	""" 
	A class used to hold various util mthods for working with strings.

	...

	Methods
	-------
	get_current_timestamp(return_as_dict, dict_key)
		Returns the current timestamp in unix time
	"""

	@staticmethod
	def get_current_timestamp(return_as_dict=True, dict_key='Timestamp'):
		"""Returns the current timestamp in unix time.

		Parameters
		----------
		return_as_dict : bool
			Wether to return the timestamp as a dictionary or as a value (default True)
		dict_key : str
			If the result is returned as a dictionary, this is the key used
		"""

		secondsSinceEpoch = round(time.time())
		if return_as_dict:
			return {dict_key: secondsSinceEpoch}
		else:
			return secondsSinceEpoch

	@staticmethod
	def get_timestamp_months_in_future(reference_timestamp, number_of_months):
		"""Returns the timestamp a certain number of months in the future from the reference timestamp.

		Parameters
		----------
		reference_timestamp: float
			The timestamp from which two months should be computed
		number_of_months: int
			The number of months in the future the timestamp returned should be. A month is a 31 day period.
		"""
		return reference_timestamp + (number_of_months * 2678400)



from NetworkStatisticsService.exceptions import CommandOutputFormatException, LatencyStatisticsException

class LatencyStatisticUtils(object):
	""" 
	A class used to extract latency statistics from ping command output

	...

	Methods
	-------
	extract_latency_statistics_from_command_output(statistics_rows)
		returns statistics dictionary or raises an error
	"""

	@staticmethod
	def __get_packet_loss_from_command_output(packet_loss_row):
		"""Extracts packet loss information from ping output row containing it.

		Parameters
		----------
		packet_loss_row : str
			The row from the ping command output containing the packet loss percentage

		Raises
		------
		CommandOutputFormatException
			If the row has unexpected format.
		"""

		stats = {}
		try:
			packet_loss_comma_split = packet_loss_row.split(",")
			for info in packet_loss_comma_split:
				if "packet loss" in info:
					stats['Loss'] = float(info.strip().split()[0][:-1])
					return stats
			raise CommandOutputFormatException("Could not find packet loss information in latency statistics.")
		except (ValueError) as error:
			raise CommandOutputFormatException("Ping command output for the packet loss row was not of expected format.\nUnderlaying error: " + str(error))

	@staticmethod
	def __get_ping_statistics_from_command_output(min_avg_max_row):
		"""Extracts latency statistics from ping command output row containing it.

		Parameters
		----------
		min_avg_max_row : str
			The row from the ping command output containing the network stats

		Raises
		------
		CommandOutputFormatException
			If the row has unexpected format.
		"""

		if min_avg_max_row is None:
			return {}
			
		try:
			elems = min_avg_max_row.split("=")
			keys = elems[0].strip().split()[-1].split("/")
			values = elems[1].strip().split()[0].split("/")
			stats = {}
			for index, key in enumerate(keys):
				stats[key.capitalize()] = float(values[index])
			return stats
		except IndexError as ie:
			raise CommandOutputFormatException("Ping command output for the packe loss row was not of expected format.\nUnderlaying error: " + str(ie))

	@staticmethod
	def extract_latency_statistics_from_command_output(statistics_rows):
		"""Extracts latency statistics from ping command output rows.

		Parameters
		----------
		statistics_rows : [str]
			The rows from the ping command output containing the network stats

		Raises
		------
		LatencyStatisticsException
			If any of the rows has unexpected format
		"""

		try:
			if "www.google.com" in statistics_rows[0]:
				packet_loss_row = statistics_rows[1]
				min_avg_max_row = None
			else:
				packet_loss_row, min_avg_max_row = statistics_rows[0], statistics_rows[1]
			loss_stats = LatencyStatisticUtils.__get_packet_loss_from_command_output(packet_loss_row)
			ping_stats = LatencyStatisticUtils.__get_ping_statistics_from_command_output(min_avg_max_row)
			return {**loss_stats, **ping_stats}
		except IndexError as ie:
			raise LatencyStatisticsException("Accessing the two statistic rows failed.\nUnderlying error: " + str(ie))
		except CommandOutputFormatException as cofe:
			raise LatencyStatisticsException("An error occured when processing the latency statistics.\nUnderlying error: " + str(cofe))



import setuptools
import os

with open("README.md", "r") as fh:
    long_description = fh.read()

if os.environ.get('CI_COMMIT_TAG'):
    version = os.environ['CI_COMMIT_TAG']
else:
    version = "local"

setuptools.setup(
    name="network-statistics-service",
    version=version,
    author="Stefan Georgescu",
    author_email="contact@stefangeorgescu.com",
	license="MIT",
   	url="https://gitlab.com/stefan.georgescu/network-statistics-service",
    description="Package for generating network statistics such as latency and download speed.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=["tests"]),
    install_requires=['speedtest-cli'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
)



import boto3
from botocore.exceptions import ClientError
import json
from decimal import Decimal
from NetworkStatisticsService.informal_statistics_uploader import InformalStatisticUploaderInterface
from NetworkStatisticsService.exceptions import DynamoDbException
from NetworkStatisticsService.time_utils import TimeUtils


class DynamoDBStatisticsUploader(InformalStatisticUploaderInterface):
	""" 
	A class used to upload to DynamoDB Tables.

	...

	Attributes
	----------
	aws_profile : str
		The AWS profile from the credential files to use (default 'default')
	aws_region : str
		The AWS region where the DynamoDB tables reside
	stats_table : str
		The name of the table where your statistics should be uplaoded
	logs_table : str
		The name of the table where error logs should be uplaoded
	ttl_config : TtlConfig
		Time to live config for DynamoDB tables.

	Methods
	-------
	upload_statistics(stats_dict)
		Uploads statistics to DynamoDB
	upload_logs(logs_dict)
		Upload error logs to DynamoDB
	"""
	def __init__(self, aws_config):
		"""
		Parameters
		----------
		aws_config : DynamoDBConfig
			DynamoDB Config object
		"""

		super().__init__()
		self.__aws_region = aws_config.aws_region
		self.__aws_profile = aws_config.aws_profile
		self.__stats_table = aws_config.stats_table
		self.__logs_table = aws_config.logs_table
		self.__ttl_config = aws_config.ttl_config
	
	def upload_statistics(self, stats_dict: dict):
		"""Uploads statistics to DynamoDB.

		If the upload fails, this method invokes upload_logs.

		Parameters
		----------
		stats_dict : dict
			Dictionary containing the statistics; must have the 'Timestamp' key
		"""

		if self.__ttl_config is not None:
			stats_dict[self.__ttl_config.attribute_name] = TimeUtils.get_timestamp_months_in_future(stats_dict["Timestamp"], self.__ttl_config.time_in_months)

		session = boto3.Session(profile_name=self.__aws_profile)
		dynamodb_client = session.resource('dynamodb', region_name=self.__aws_region)
		table = dynamodb_client.Table(self.__stats_table)
		stats_dict = json.loads(json.dumps(stats_dict), parse_float=Decimal)
		
		try:
			table.put_item(Item=stats_dict)
		except ClientError as ce:
			error_log = {"Error": ce, "Timestamp": stats_dict["Timestamp"]}
			self.upload_logs(error_log)

	def upload_logs(self, logs_dict: dict):
		"""Upload error logs to DynamoDB.

		Parameters
		----------
		logs_dict : dict
			Dictionary containing the error; must have the 'Timestamp' key

		Raises
		------
		DynamoDbException
			If upload of error dictionary fails.
		"""

		session = boto3.Session(profile_name=self.__aws_profile)
		dynamodb_client = session.resource('dynamodb', region_name=self.__aws_region)
		table = dynamodb_client.Table(self.__logs_table)
		stats_dict = json.loads(json.dumps(stats_dict), parse_float=Decimal)
		
		try:
			table.put_item(Item=stats_dict)
		except ClientError as ce:
			raise DynamoDbException("Failed to upload logs to DynamoDB: " + ce)
